#include <stdio.h>
#include <inttypes.h>
#include <endian.h>

#include "eth_raw.h"
#include "error.h"
#include "eth_hw.h"


#if !defined(__BYTE_ORDER) || __BYTE_ORDER != __LITTLE_ENDIAN
	#error "Endianess nicht definiert."
#endif

#define htonl(x) __builtin_bswap32(x)
#define htons(x) __builtin_bswap16(x)
#define ntohl(x) __builtin_bswap32(x)
#define ntohs(x) __builtin_bswap16(x)

int32_t counter = 0;
static uint32_t ip_device;
static uint8_t mac_device[6];
static uint8_t mac_broadcast[6] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };


struct connection_s
	{
	uint32_t ip;
	uint8_t mac[6];
	uint16_t port;
	};

struct connection_s connections[20] = {0};
		

struct ethheader_s
	{
	uint8_t dest[6];
	uint8_t src[6];
	uint16_t ethertype;
	uint8_t payload[];
	} __attribute__ ((__packed__));

struct ipheader_s
	{
#if __BYTE_ORDER == __LITTLE_ENDIAN
    uint8_t ihl:4;
    uint8_t version:4;
#elif __BYTE_ORDER == __BIG_ENDIAN
    uint8_t version:4;
    uint8_t ihl:4;
#else
# error "Unbekannte endianess"
#endif
	uint8_t tos;
	uint16_t totallength;
	uint16_t identification;
	uint16_t fragment_offset;
	uint8_t ttl;
	uint8_t protocol;
	uint16_t header_checksum;
	uint32_t src;
	uint32_t dst;
	uint8_t payload[];
	} __attribute__ ((__packed__));

struct arp_s
	{
	uint16_t hwaddrtype;
	uint16_t protoaddrtype;
	uint8_t hwaddrsize;
	uint8_t protoaddrsize;
	uint16_t operation;		// 1 = Anfrage, 2 = Antwort
	uint8_t mac_src[6];
	uint32_t ip_src;
	uint8_t mac_dest[6];
	uint32_t ip_dest;
	} __attribute__ ((__packed__));

struct icmp_echo_s
	{
	uint8_t type;
	uint8_t code;
	uint16_t checksum;
	uint16_t identifier;
	uint16_t sequence_number;
	uint8_t payload[];
	} __attribute__ ((__packed__));

struct tcpheader_s
	{
	
	} __attribute__ ((__packed__));

void print_mac(uint8_t mac[6])
	{
	printf("%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx\n",
		mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	}

void print_ip(uint32_t ip)
	{
	printf("%hhu.%hhu.%hhu.%hhu\n",
		((uint8_t*)&ip)[0],((uint8_t*)&ip)[1], ((uint8_t*)&ip)[2],((uint8_t*)&ip)[3]);
	}


static inline int mac_equals(const uint8_t a[6], const uint8_t b[6])
	{
	return
        ( *(uint32_t*)&(a[0]) == *(uint32_t*)&(b[0]) ) &
        ( *(uint16_t*)&(a[4]) == *(uint16_t*)&(b[4]) );
	}


void eth_init(uint32_t ip, uint8_t mac[6])
	{
	ip_device = ip;
	memcpy(mac_device, mac, sizeof(mac_device));
	eth_hw_init();
	}


void eth_uninit(void)
	{
	eth_hw_uninit();
	}


void handle_icmp(void *data, int length, struct ethheader_s *ethframe)
	{
	printf("ICMP\n");
	if(length >= sizeof(struct icmp_echo_s))
		{
		struct icmp_echo_s *echo = (struct icmp_echo_s *)data;
		if(echo->type == 8 && echo->code == 0)
			{
			struct ethheader_s *ethheader = eth_hw_alloc();
			if(ethheader != NULL)
				{
				memcpy(ethheader->dest, ethframe->src, sizeof(ethheader->dest));
				memcpy(ethheader->src, mac_device, sizeof(ethheader->src));
				ethheader->ethertype = htons(0x0800);
				
				struct ipheader_s *iph = (struct ipheader_s *)&(ethheader->payload);
				iph->version = 4;
				iph->ihl = 5;
				iph->tos = 0;
				iph->totallength = htons(sizeof(struct ipheader_s) + length);
				iph->identification = htons(ntohs(((struct ipheader_s *)&(ethframe->payload))->identification)+1);
				iph->fragment_offset = 0x40;
				iph->ttl = 64;
				iph->protocol = 1;
				uint16_t *hd = (uint16_t*)&(ethheader->payload);
				iph->header_checksum = ~(hd[0] + hd[1] + hd[2] + hd[3] + hd[4] /*+ hd[5]*/ + hd[6] + hd[7] + hd[8] + hd[9]); // SAM macht das automatisch.
				iph->src = ip_device;
				iph->dst = ((struct ipheader_s *)&(ethframe->payload))->src;
				
				struct icmp_echo_s *reply = (struct icmp_echo_s *)&(iph->payload);
				reply->type = 0;
				reply->code = 0;
				reply->identifier = echo->identifier;
				reply->sequence_number = echo->sequence_number;
				memcpy(reply->payload, echo->payload, length-sizeof(*reply));
				reply->checksum = echo->checksum + (uint16_t)8;
				
				eth_hw_send(ethheader, sizeof(struct ethheader_s) + sizeof(struct ipheader_s) + length);
				printf("Ping");
				}
			}
		else
			{
			ERROR_APPEND("Ungültige ICMP Nachricht, typ = %hhu, code = %hhu", echo->type, echo->code);
			}
		}
	else
		{
		ERROR_APPEND("Zu kurze ICMP Nachricht.");
		}
	
	eth_hw_free_offset(data);
	}


void handle_ip(void *data, int length, struct ethheader_s *ethframe)
	{
	if(length <= sizeof(struct ipheader_s))
		{
		ERROR_APPEND("Zu kurzes IP Paket.");
		eth_hw_free(data);
		return;
		}
	struct ipheader_s *ipframe = (struct ipheader_s *)data;
	printf("IP Source: "); print_ip(ipframe->src);
	printf("IP Destination: "); print_ip(ipframe->dst);

	int headerlen = (int)ipframe->ihl *4;
	if(	ipframe->version != 4 || 					// Kein IPv4
		headerlen < 20 ||							// Ungültige Headerlänge
		headerlen >= length ||
		ntohs(ipframe->totallength) > length || 	// Paket ist Fragmentiert
		(ipframe->fragment_offset != ntohs(0x4000) && ipframe->fragment_offset != 0)) 	// Paket ist Fragmentiert
		{
		ERROR_APPEND("Ungültiges IP Paket.");
		eth_hw_free_offset(data);
		return;
		}

	int protokoll = ipframe->protocol;
	void *payload = (void*)ipframe + headerlen;
	int payloadlength = length - headerlen;
	if(protokoll == 1)	// ICMP
		{
		handle_icmp(payload, payloadlength, ethframe);
		}
	else if(protokoll == 6)	// TCP
		{
#warning "TODO"
		}
	else	// Unbekanntes Protokoll
		{
		eth_hw_free_offset(data);
		}
	}


void handle_arp(void *data, int length)
	{
	if(length < sizeof(struct arp_s))
		{
		ERROR_APPEND("Zu kurzes ARP Paket.");
		eth_hw_free_offset(data);
		return;
		}
	struct arp_s *arp = (struct arp_s *)data;
	if(arp->hwaddrtype == htons(1))
		{
		if(arp->protoaddrtype == htons(0x0800))
			{
			if(arp->hwaddrsize == 6)
				{
				if(arp->protoaddrsize == 4)
					{
					if(arp->operation == htons(1))
						{
						if(arp->ip_dest == ip_device)
							{
							printf("ARP Anfrage von %hhu.%hhu.%hhu.%hhu\n", ((uint8_t*)&(arp->ip_dest))[0], ((uint8_t*)&(arp->ip_dest))[1], ((uint8_t*)&(arp->ip_dest))[2], ((uint8_t*)&(arp->ip_dest))[3]);
							struct ethheader_s *sendframe = eth_hw_alloc();
							if(sendframe != NULL)
								{
								_Static_assert(sizeof(sendframe->dest) == sizeof(arp->mac_src), "Unterschiedliche größen für MAC Adresse.");
								memcpy(sendframe->dest, arp->mac_src, sizeof(sendframe->dest));
								memcpy(sendframe->src, mac_device, sizeof(sendframe->src));
								sendframe->ethertype = htons(0x0806);

								struct arp_s *sendarp = (struct arp_s *)&(sendframe->payload);
								sendarp->hwaddrtype = htons(0x0001);
								sendarp->protoaddrtype = htons(0x0800);
								sendarp->hwaddrsize = 6;
								sendarp->protoaddrsize = 4;
								sendarp->operation = htons(2);
								memcpy(&(sendarp->mac_src), mac_device, sizeof(sendarp->mac_src));
								sendarp->ip_src = ip_device;
								memcpy(&(sendarp->mac_dest), arp->mac_src, sizeof(sendarp->mac_src));
								sendarp->ip_dest = arp->ip_src;

								printf("Ping reply an: "); print_ip(sendarp->ip_dest);
								eth_hw_send(sendframe, sizeof(struct ethheader_s) + sizeof(struct arp_s));
								}
							else
								{
								// Kein Speicher frei, Paket ignorieren.
								ERROR_APPEND("Speicher voll.");
								}
							}
						else
							{
							// Anfrage für anderen Empfänger
							ERROR_APPEND("ARP für anderen Empfänger, %hhu.%hhu.%hhu.%hhu ist %hhu.%hhu.%hhu.%hhu\n",
								((uint8_t*)&(arp->ip_dest))[0], ((uint8_t*)&(arp->ip_dest))[1], ((uint8_t*)&(arp->ip_dest))[2], ((uint8_t*)&(arp->ip_dest))[3],
								((uint8_t*)&(ip_device))[0], ((uint8_t*)&(ip_device))[1], ((uint8_t*)&(ip_device))[2], ((uint8_t*)&(ip_device))[3]);
							}
						}
					else
						{
						ERROR_APPEND("Ungültiger ARP Operationscode %hu.\n", ntohs(arp->operation));
						}
					}
				else
					{
					ERROR_APPEND("Ungültige ARP Protokolladresslänge %hhu.\n", ntohs(arp->protoaddrsize));
					}
				}
			else
				{
				ERROR_APPEND("Ungültige ARP Hardwareadresslänge %hhu.\n", ntohs(arp->hwaddrsize));
				}
			}
		else
			{
			ERROR_APPEND("Unbekannter ARP Protokolladresstyp %hu.\n", ntohs(arp->protoaddrtype));
			}
		}
	else
		{
		ERROR_APPEND("Unbekannter ARP Hardwareadresstyp %hu.\n", ntohs(arp->hwaddrtype));
		}
	eth_hw_free_offset(data);
	}


void handle_ethernet(void *data, int length)
	{
	if(length <= sizeof(struct ethheader_s))
		{
		ERROR_APPEND("Zu kurzes Ethernet Paket.\n");
		eth_hw_free(data);
		return;
		}
	struct ethheader_s *ethframe = data;
	printf("MAC Destination: "); print_mac(ethframe->dest);
	printf("MAC Source: "); print_mac(ethframe->src);

	if(	!mac_equals(ethframe->dest, mac_device) &&
		!mac_equals(ethframe->dest, mac_broadcast))
		{
		// Paket ist nicht für uns
		ERROR_APPEND("Paket ist nicht für uns.");
		eth_hw_free(data);
		return;
		}

	if(ethframe->ethertype == htons(0x0800))	// IP?
		{
		// ggf Paket mit Broadcast mac verwerfen wenn IP
		handle_ip(ethframe->payload, length - sizeof(*ethframe), ethframe);
		}
	else if(ethframe->ethertype == htons(0x0806)) // ARP?
		{
		handle_arp(ethframe->payload, length - sizeof(*ethframe));
		}
	else
		{
		printf("Unbekannter Ethertype %04hX.\n", ntohs(ethframe->ethertype));
		eth_hw_free(data);
		return;
		}
	}


void eth_poll(void)
	{
	for(;;)
		{
		void *buffer;
		int length = eth_hw_read(&buffer);
		if(length > 0)
			{
			counter++;
			puts("");
			printf("Paket %"PRId32", %d bytes read\n", counter, length);
			handle_ethernet(buffer, length);
			}
		}
	}
